@@ -91,7 +91,7 @@ soup_uri_protocol_to_string (SoupProtocol proto)
 	return "";
 }
 
-gint
+static gint
 soup_uri_get_default_port (SoupProtocol proto)
 {
 	SoupKnownProtocols *known = known_protocols;
@@ -330,8 +330,9 @@ soup_uri_new (const gchar* uri_string)
 	/* If there is an @ sign, look for user, authmech, and
 	 * password before it.
 	 */
+	slash = strchr (uri_string, '/');
 	at = strchr (uri_string, '@');
-	if (at) {
+	if (at && (!slash || at < slash)) {
 		colon = strchr (uri_string, ':');
 		if (colon && colon < at)
 			g_uri->passwd = g_strndup (colon + 1, at - colon - 1);
@@ -355,7 +356,6 @@ soup_uri_new (const gchar* uri_string)
 		g_uri->user = g_uri->passwd = g_uri->authmech = NULL;
 
 	/* Find host (required) and port. */
-	slash = strchr (uri_string, '/');
 	colon = strchr (uri_string, ':');
 	if (slash && colon > slash)
 		colon = 0;
@@ -392,7 +392,6 @@ soup_uri_new (const gchar* uri_string)
 	if (path && query) {
 		g_uri->path = g_strndup (path, query - path);
 		g_uri->querystring = g_strdup (++query);
-		g_uri->query_elems = g_strsplit (g_uri->querystring, "&", 0);
 		g_free (path);
 	} else {
 		g_uri->path = path;
@@ -414,7 +413,7 @@ soup_uri_to_string (const SoupUri *uri, gboolean show_passwd)
 	if (uri->port != -1 &&
 	    uri->port != soup_uri_get_default_port (uri->protocol))
 		return g_strdup_printf(
-			"%s%s%s%s%s%s%s%s:%d%s%s%s",
+			"%s%s%s%s%s%s%s%s:%d%s%s%s%s",
 			soup_uri_protocol_to_string (uri->protocol),
 			uri->user ? uri->user : "",
 			uri->authmech ? ";auth=" : "",
@@ -424,12 +423,13 @@ soup_uri_to_string (const SoupUri *uri, gboolean show_passwd)
 			uri->user ? "@" : "",
 			uri->host,
 			uri->port,
+			uri->path && *uri->path != '/' ? "/" : "",
 			uri->path ? uri->path : "",
 			uri->querystring ? "?" : "",
 			uri->querystring ? uri->querystring : "");
 	else
 		return g_strdup_printf(
-			"%s%s%s%s%s%s%s%s%s%s%s",
+			"%s%s%s%s%s%s%s%s%s%s%s%s",
 			soup_uri_protocol_to_string (uri->protocol),
 			uri->user ? uri->user : "",
 			uri->authmech ? ";auth=" : "",
@@ -438,31 +438,7 @@ soup_uri_to_string (const SoupUri *uri, gboolean show_passwd)
 			uri->passwd && show_passwd ? uri->passwd : "",
 			uri->user ? "@" : "",
 			uri->host,
-			uri->path ? uri->path : "",
-			uri->querystring ? "?" : "",
-			uri->querystring ? uri->querystring : "");
-}
-
-gchar *
-soup_uri_to_proxyable_string (const SoupUri *uri)
-{
-	g_return_val_if_fail (uri != NULL, NULL);
-
-	if (uri->port != -1 &&
-	    uri->port != soup_uri_get_default_port (uri->protocol))
-		return g_strdup_printf(
-			"%s%s:%d%s%s%s",
-			soup_uri_protocol_to_string (uri->protocol),
-			uri->host,
-			uri->port,
-			uri->path ? uri->path : "",
-			uri->querystring ? "?" : "",
-			uri->querystring ? uri->querystring : "");
-	else
-		return g_strdup_printf(
-			"%s%s%s%s%s",
-			soup_uri_protocol_to_string (uri->protocol),
-			uri->host,
+			uri->path && *uri->path != '/' ? "/" : "",
 			uri->path ? uri->path : "",
 			uri->querystring ? "?" : "",
 			uri->querystring ? uri->querystring : "");
@@ -471,22 +447,44 @@ soup_uri_to_proxyable_string (const SoupUri *uri)
 SoupUri *
 soup_uri_copy (const SoupUri* uri)
 {
-	gchar *uri_str;
 	SoupUri *dup;
 
 	g_return_val_if_fail (uri != NULL, NULL);
 
-	uri_str = soup_uri_to_string (uri, TRUE);
-	dup = soup_uri_new (uri_str);
-	g_free (uri_str);
+	dup = g_new0 (SoupUri, 1);
+	dup->protocol    = uri->protocol;
+	dup->user        = g_strdup (uri->user);
+	dup->authmech    = g_strdup (uri->authmech);
+	dup->passwd      = g_strdup (uri->passwd);
+	dup->host        = g_strdup (uri->host);
+	dup->port        = uri->port;
+	dup->path        = g_strdup (uri->path);
+	dup->querystring = g_strdup (uri->querystring);
 
 	return dup;
 }
 
+gboolean 
+soup_uri_equal (const SoupUri *u1, 
+		const SoupUri *u2)
+{
+	if (u1->protocol == u2->protocol            &&
+	    u1->port     == u2->port                &&
+	    !strcmp (u1->user,        u2->user)     &&
+	    !strcmp (u1->authmech,    u2->authmech) &&
+	    !strcmp (u1->passwd,      u2->passwd)   &&
+	    !strcmp (u1->host,        u2->host)     &&
+	    !strcmp (u1->path,        u2->path)     &&
+	    !strcmp (u1->querystring, u2->querystring))
+		return TRUE;
+
+	return FALSE;
+}
+
 void
 soup_uri_free (SoupUri *uri)
 {
-	g_assert (uri);
+	g_return_if_fail (uri != NULL);
 
 	g_free (uri->user);
 	g_free (uri->authmech);
@@ -494,11 +492,27 @@ soup_uri_free (SoupUri *uri)
 	g_free (uri->host);
 	g_free (uri->path);
 	g_free (uri->querystring);
-	g_strfreev (uri->query_elems);
 
 	g_free (uri);
 }
 
+void
+soup_uri_set_auth  (SoupUri       *uri, 
+		    const gchar   *user, 
+		    const gchar   *passwd, 
+		    const gchar   *authmech)
+{
+	g_return_if_fail (uri != NULL);
+
+	g_free (uri->user);
+	g_free (uri->passwd);
+	g_free (uri->authmech);
+
+	uri->user = g_strdup (user);
+	uri->passwd = g_strdup (passwd);
+	uri->authmech = g_strdup (authmech);
+}
+
 void
 soup_debug_print_uri (SoupUri *uri)
 {
